#!/bin/bash

# Exit on error
set -e

# Require root or sudo
if [[ $EUID -ne 0 ]]; then
   echo "This script should must be run using sudo or as the root user" 1>&2
   exit 1
fi

# Portions of this code are based upon the BSFL:
# ###########################
# Bash Shell Function Library
# ###########################
#
# Author: Louwrentius <louwrentius@gmail.com>
# Maintainer: Peter Boling <peter.boling@gmail.com>
# Contributions by: Jani Hurskainen
#
# Copyright © 2010, 2013, 2014
#
# Released under the current GPL version.

echo_failure() {
  printf "\e[31m ✘ ${1} \e[0m"
}

echo_success() {
  printf "\e[32m ✔ ${1} \e[0m"
}

step() {
    echo -n "$@"

    STEP_OK=0
    [[ -w /tmp ]] && echo $STEP_OK > /tmp/step.$$
}

try() {
    local BG=1;

    [[ $1 == -- ]] && {       shift; }

    # Run the command.
    "$@"

    # Check if command failed and update $STEP_OK if so.
    local EXIT_CODE=$?

    if [[ $EXIT_CODE -ne 0 ]]; then
        STEP_OK=$EXIT_CODE
        [[ -w /tmp ]] && echo $STEP_OK > /tmp/step.$$

        if [[ -n $LOG_STEPS ]]; then
            local FILE=$(readlink -m "${BASH_SOURCE[1]}")
            local LINE=${BASH_LINENO[0]}

            echo "$FILE: line $LINE: Command \`$*' failed with exit code $EXIT_CODE." >> "$LOG_STEPS"
        fi
    fi

    return $EXIT_CODE
}

next() {
    [[ -f /tmp/step.$$ ]] && { STEP_OK=$(< /tmp/step.$$); rm -f /tmp/step.$$; }
    [[ $STEP_OK -eq 0 ]]  && echo_success || echo_failure
    echo

    return $STEP_OK
}

is_installed() {
    dpkg-query -Wf'${db:Status-abbrev}' "$1" 2>/dev/null | grep -q '^i'
}

command_exists () {
    type "$1" &> /dev/null ;
}

pg_extension_exists () {
    if [ -f "/usr/lib/postgresql/$PG_VERSION/lib/$1.so" ]; then
        return 0
    fi

    if [ -f "/usr/share/postgresql/$PG_VERSION/extension/$1.control" ]; then
        return 0
    fi

    return 1
}

psql_wrapper () {
    local SQL=$1
    
    if [ -n "CI" ]; then
        su postgres -c "psql -c \"$1"\""
    else
        psql -c "$1""
    fi
}

pgxn_install () {
    local file=$1
    local file=${2:-$file}

    export USE_PGXS=1

    if ! pg_extension_exists ${file}
    then
        echo "$file... installing"
        if [ -z "$3" ]; then
            pgxn install --yes $1
        else
            pgxn install --yes --$3 $1
        fi
    else
        return 0
    fi
}

pgxn_install_unstable () {
    local file=$1
    local file=${2:-$file}

    export USE_PGXS=1

    if ! pg_extension_exists ${file}
    then
        echo "$file... installing"
        if [ -z "$3" ]; then
            pgxn install --unstable --yes $1
        else
            pgxn install --unstable --yes --$3 $1
        fi
    else
        return 0
    fi
}

locale_set=$(locale | grep en_US.UTF-8)

if [ ! locale_set ]; then
    step "Setting locale: "
        for y in $(locale | cut -d '=' -f 2| sort |uniq ); do locale-gen $y; done
        try locale-gen "en_US.UTF-8"
        try dpkg-reconfigure locales
        try update-locale LANG=en_US.UTF-8
    next
fi

if ! command_exists "add-apt-repository" ] || ! is_installed "git" || ! is_installed "unzip" || ! is_installed "build-essential" || ! is_installed "pkg-config" || ! is_installed "libtool"
then
    step "Installing packages required to install other packages: "
        try apt-get update
        try apt-get upgrade -y
        try apt-get install -y dialog python-software-properties python-setuptools wget curl unzip nano ca-certificates git sudo build-essential pkg-config autoconf libtool cmake
        try git config --global core.editor "nano"
    next
fi

if ! is_installed "postgresql-server-dev-$PG_VERSION"
then
    wget --quiet -O - https://www.postgresql.org/media/keys/ACCC4CF8.asc | sudo apt-key add -

    step "Installing PostgreSQL $PG_VERSION: "
    if [ "$PG_VERSION" = "9.4" ]; then
        try echo "deb http://apt.postgresql.org/pub/repos/apt/ $(lsb_release -cs)-pgdg main" > /etc/apt/sources.list.d/pgdg.list
        try apt-get update
        try apt-get install -y postgresql postgresql-client postgresql-contrib libpq-dev libpq5 postgresql-server-dev-$PG_VERSION postgresql-plperl-$PG_VERSION pgagent
    else
        try echo "deb http://apt.postgresql.org/pub/repos/apt/ $(lsb_release -cs)-pgdg main $PG_VERSION" > /etc/apt/sources.list.d/pgdg.list
        try apt-get update
        try apt-get install -y postgresql-$PG_VERSION postgresql-client-$PG_VERSION  postgresql-contrib-$PG_VERSION postgresql-doc-$PG_VERSION postgresql-plperl-$PG_VERSION postgresql-plpython-$PG_VERSION postgresql-server-dev-$PG_VERSION libpgtypes3 pgagent
    fi
    next
fi

if ! command_exists "pgxn"
then
    step "Installing PGXN: "
        try easy_install pgxnclient
    next
fi

if ! is_installed "libv8-dev"
then
    step "Installing development dependencies for postgres extensions: "
        try apt-get install -y libyajl-dev libmysqld-dev protobuf-c-compiler libprotobuf-c0-dev libv8-dev
    next
fi

if ! is_installed "postgresql"
then
    step "Installing postgresql extensions using PGXN: "
        try pgxn_install_unstable "jsoncdc"
        try pgxn_install "plv8"
    next
fi

if [[ $(sudo -u postgres psql -lqt | cut -d \| -f 1 | grep -w lapidus | wc -l) == 0 ]]; then
    step "Creating template database with proper encoding: "
        try su postgres -c "psql -c \"UPDATE pg_database SET datistemplate = FALSE WHERE datname = 'template1';\""
        try su postgres -c "psql -c \"DROP DATABASE template1;\""
        try su postgres -c "psql -c \"CREATE DATABASE  template1 with ENCODING = 'UTF-8' LC_CTYPE = 'en_US.utf8' LC_COLLATE = 'en_US.utf8' template = template0;\""
        try su postgres -c "psql -c \"UPDATE pg_database SET datistemplate = TRUE WHERE datname = 'template1';\""
    next

    step "Creating roles and databases in PostgreSQL for testing"
        try sudo -u postgres psql -c "CREATE USER lapidus SUPERUSER REPLICATION LOGIN ENCRYPTED PASSWORD '';"
        try sudo -u postgres createdb lapidus -O lapidus
        try sudo -u postgres psql -c "CREATE USER hurley SUPERUSER REPLICATION LOGIN ENCRYPTED PASSWORD '';"
        try sudo -u postgres createdb hurley -O hurley
    next
fi

if ! is_installed nodejs || ! command_exists npm
then
    step "Installing node.js 4.X"
        try apt-get -y remove nodejs
        try curl -sL https://deb.nodesource.com/setup_4.x | sudo bash -
        try apt-get install -y nodejs
    next
fi

mysql -e "CREATE DATABASE jacob;" -uroot
mysql -e "CREATE DATABASE sawyer;" -uroot
mysql -u root jacob < test/sql/test_table.sql
mysql -u root sawyer < test/sql/test_table.sql
mysql -e "CREATE USER 'jacob'@'localhost' IDENTIFIED BY '2PQM9aiKMJX5chv76gYdFJNi';" -uroot
mysql -e "CREATE USER 'sawyer'@'localhost' IDENTIFIED BY '2PQM9aiKMJX5chv76gYdFJNi';" -uroot
mysql -e "GRANT SELECT, REPLICATION SLAVE, REPLICATION CLIENT ON *.* TO 'jacob'@'localhost';" -uroot
mysql -e "GRANT SELECT, REPLICATION SLAVE, REPLICATION CLIENT ON *.* TO 'sawyer'@'localhost';" -uroot
mysql -e "FLUSH PRIVILEGES;" -uroot

psql_wrapper("CREATE ROLE lapidus PASSWORD '2PQM9aiKMJX5chv76gYdFJNi' SUPERUSER CREATEDB CREATEROLE INHERIT LOGIN;");
psql_wrapper('CREATE DATABASE lapidus OWNER lapidus;');
psql_wrapper("CREATE ROLE hurley PASSWORD '2PQM9aiKMJX5chv76gYdFJNi' SUPERUSER CREATEDB CREATEROLE INHERIT LOGIN;");
psql_wrapper('CREATE DATABASE hurley OWNER hurley;');

mkdir the_hatch
mongod --port 27017 --dbpath the_hatch --replSet rs0 --oplogSize 20 --noprealloc --fork --smallfiles --logpath mongodb.log
sleep 5
mongo lapidus --eval 'printjson(rs.initiate());'
sleep 20
mongo lapidus --eval 'db.getMongo().setSlaveOk();'
mongo lapidus --eval 'db.createUser({ user: "lapidus", pwd: "2PQM9aiKMJX5chv76gYdFJNi", roles: [ {role: "clusterAdmin", db: "admin" }, { role: "readAnyDatabase", db: "admin" }, "readWrite" ] });'
